"
 COPYRIGHT (c) 1997 by Claus Gittinger / eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"




Object subclass:#DataSetColumn
	instanceVariableNames:'columnNumber dataSet minWidth width description buttonExtent
		drawableAction rendererType backgroundColor rowSeparatorSelector
		showColSeparator showRowSeparator foregroundColor fgSelector
		bgSelector columnAlignment label readSelector
		numArgsToReadSelector columnAdaptor'
	classVariableNames:''
	poolDictionaries:''
	category:'Views-DataSet'
!

!DataSetColumn class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1997 by Claus Gittinger / eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"



!

documentation
"
    represent one single column description of a DataSetView

    [Instance variables:]

        columnNumber    <Integer>               sequence number (into list of columns)

        dataSet         <DSVColumnView>         the view it belongs to

        minWidth        <Integer>               minimum required width by the column

        width           <Integer>               width of column

        description     <DataSetColumnSpec>     the column description

        buttonExtent    <Point>                 extent of drawable form

        drawableAction  <Action>                action to access the printable
                                                label of a cell

        rendererType    <Type>                  renderer type (cached value from
                                                the column specification).

        backgroundColor <Color or nil>          background color of all cells or nil
                                                (nil: use default background color).

        foregroundColor <Color or nil>          foreground color of all cells or nil
                                                (nil: use default foreground color).

        fgSelector      <Selector or nil>       access specific foreground color for
                                                a cell

        bgSelector      <Selector or nil>       access specific background color for
                                                a cell

        columnAlignment <Symbol>                align text/icon #left #right or #center
                                                in row (on default: #left).

        label           <label/icon or nil>     label resolved by the builder shown
                                                in the column description field.

        readSelector    <Symbol>                cached readSelector (from the spec)

        numArgsToReadSelector <Integer>         number of arguments to the readSelector

        columnAdaptor   <nil or any>            if non-nil, that one is asked (via
                                                read-writeSelectors to extract a column
                                                from a row object.
                                                If nil, the row object is used itself.

        showColSeparator <Boolean>              true: vertical separators are enabled;
                                                separator between columns

        showRowSeparator <Boolean>              true: horizontal separators are enabled;
                                                separator between rows

        rowSeparatorSelector <Selector or nil>  access specific showRowSeparator state
                                                for a cell

    [author:]
        Claus Atzkern

    [see also:]
        DataSetColumnSpec
        DSVColumnView
        DataSetView
"


! !

!DataSetColumn methodsFor:'accessing'!

at:aRowNr
    "get the value of the raw at an index, aRowNr
    "
    ^ self extractColFromRow:(dataSet at:aRowNr).
!

at:aRowNr put:something
    "set the value of the raw at an index, aRowNr
    "
    |row newRow|

    row := dataSet at:aRowNr.
    newRow := self storeCol:something inRow:row.
    newRow ~~ row ifTrue:[
        dataSet dataSetView listAt:aRowNr put:newRow.
    ]

    "Modified: / 7.8.1998 / 22:17:27 / cg"
!

backgroundColor
    "returns the background color or nil
    "
    ^ backgroundColor == dataSet backgroundColor ifTrue:[nil]
                                                ifFalse:[backgroundColor]
!

description
    "returns the column description
    "
    ^ description


!

foregroundColor
    "returns the foreground color or nil
    "
    ^ foregroundColor == dataSet foregroundColor ifTrue:[nil]
                                                ifFalse:[foregroundColor]
!

label
    "get the label assigned to the column
    "
    ^ label


!

rendererType
    "returns my renderer type
    "
    ^ rendererType
!

width
    "returns the width in pixels
    "
    |end|

    width notNil ifTrue:[ ^ width ].                            "/ already computed

    (    (width := description width)    ~~ 0                   "/ fixed size
     or:[(width := description minWidth) ~~ 0]                  "/ start size
    ) ifFalse:[
        self containsText ifTrue:[
            "/ take maximum 10 entries to calculate the width;
            "/ can be resized later if neccessary

            end := 10 min:dataSet size.

            1 to:end do:[:aRowNr||lbl|
                lbl := drawableAction value:(dataSet at:aRowNr).

                (lbl respondsTo:#widthOn:) ifTrue:[
                    width := width max:(lbl widthOn:dataSet)
                ] ifFalse:[
                    lbl notNil ifTrue:[
                        width := width max:(lbl displayString widthOn:dataSet)
                    ]
                ]
            ].
            description editorType ~~ #None ifTrue:[
                width := width + (dataSet font widthOn:dataSet device)
            ].
        ].
        width := width + buttonExtent x.

        label notNil ifTrue:[
            width := width max:(4 + (label widthOn:dataSet))
        ]
    ].
    ^ width := minWidth := width + dataSet separatorSize + (2 * dataSet horizontalSpacing).
!

width:aWidth
    "set the width
    "
    |w|

    (w := description width) ~~ 0 ifTrue:[             "/ fixed size
        width := minWidth := w
    ] ifFalse:[
        width := aWidth max:minWidth
    ].

! !

!DataSetColumn methodsFor:'drawing'!

drawLabelAt:aCellIndex x:xLeft y:yTop h:h
    "redraw label in cell
    "
    |fg bg label row isSel
     y "{ Class:SmallInteger }"
     x "{ Class:SmallInteger }"
    |
    row := dataSet at:aCellIndex.
    y := yTop.
    x := xLeft.

    isSel := dataSet isSelected:aCellIndex inColumn:columnNumber.

    (isSel and:[description showSelectionHighLighted]) ifTrue:[
        fg := dataSet hgLgFgColor.
        bg := dataSet hgLgBgColor.
    ] ifFalse:[
        bgSelector notNil ifTrue:[
            (bg := row perform:bgSelector) notNil ifTrue:[
                bg := dataSet colorOnDevice:bg
            ] ifFalse:[
                (bg := backgroundColor) isNil ifTrue:[
                    bg := dataSet backgroundColor
                ]
            ]
        ] ifFalse:[
            bg := nil
        ].

        (fgSelector notNil and:[(fg := row perform:fgSelector) notNil]) ifTrue:[
            fg := dataSet colorOnDevice:fg
        ] ifFalse:[
            fg := foregroundColor
        ]
    ].

    bg notNil ifTrue:[
        dataSet paint:bg.
        dataSet fillRectangleX:x y:y width:width height:h.
    ] ifFalse:[
        (bg := backgroundColor) isNil ifTrue:[
            bg := dataSet backgroundColor
        ]
    ].

    rendererType == #rowSelector ifTrue:[
        isSel ifFalse:[ ^ self ].
        label := dataSet rowSelectorForm.
        x  := x + (dataSet horizontalSpacing).
    ] ifFalse:[
        (label := drawableAction value:row) isNil ifTrue:[ ^ self ].

        columnAlignment == #left ifTrue:[
            x  := x + (dataSet horizontalSpacing).
        ] ifFalse:[
            x := width - (label widthOn:dataSet).

            columnAlignment == #right ifTrue:[x := x - (dataSet horizontalSpacing)]
                                     ifFalse:[x := x // 2].
            x := (xLeft + x) max:0.
        ]
    ].

    y := y + (h - (label heightOn:dataSet) // 2).

    label isImageOrForm ifFalse:[
        y := y + dataSet rowFontAscent
    ].
    dataSet paint:fg on:bg.
    label displayOn:dataSet x:x y:y



!

drawSeparatorsAt:aCellIndex x:xLeft y:yTop mode3D:is3D
    "redraw separators for cell
    "
    |xL    "{ Class:SmallInteger }"
     xR    "{ Class:SmallInteger }"
     yB    "{ Class:SmallInteger }"
     yT    "{ Class:SmallInteger }"
    |
    is3D ifTrue:[dataSet paint:(dataSet separatorDarkColor)]
        ifFalse:[dataSet paint:(dataSet foregroundColor)].

    xL := xLeft.
    yT := yTop.
    xR := xL - 1 + width.
    yB := yT - 1 + dataSet rowHeight.

    showRowSeparator ifTrue:[
        (    rowSeparatorSelector isNil or:[aCellIndex == dataSet size
          or:[((dataSet at:aCellIndex) perform:rowSeparatorSelector) ~~ false]]
        ) ifTrue:[
            dataSet displayLineFromX:xL y:yB toX:xR y:yB
        ]
    ].
    showColSeparator ifTrue:[
        dataSet displayLineFromX:xR y:yT toX:xR y:yB.
    ].

    is3D ifFalse:[
        ^ self
    ].

    dataSet paint:(dataSet separatorLightColor).

    (     columnNumber == 1
      or:[(dataSet columnAt:(columnNumber - 1)) showColSeparator]
    ) ifTrue:[
        dataSet displayLineFromX:xL y:yT toX:xL y:yB.
    ] ifFalse:[
        xL := xL - 1.
    ].

    showRowSeparator ifTrue:[
        (    rowSeparatorSelector isNil or:[aCellIndex == 1
          or:[((dataSet at:(aCellIndex - 1)) perform:rowSeparatorSelector) ~~ false]]
        ) ifTrue:[
            dataSet displayLineFromX:xL y:yT toX:xR - 1 y:yT.
        ]
    ].


!

invalidate
    "set to invalidate; forces a recomputation
    "
    minWidth := width := nil.

!

redrawX:xLeft y:yTop toY:yBot from:start to:stop
    "redraw rows between start and stop
    "
    |is3D bg
     h "{ Class:SmallInteger }"
     x "{ Class:SmallInteger }"
     y "{ Class:SmallInteger }"
    |
    h := dataSet rowHeight.
    y := yTop.
    x := xLeft.

    bgSelector isNil ifTrue:[
        (bg := backgroundColor) isNil ifTrue:[
            bg := dataSet backgroundColor
        ].
        dataSet paint:bg.
        dataSet fillRectangleX:x y:y width:width height:(yBot - yTop)
    ].

 "/ DRAW CELL: BACKGROUND/FOREGROUND/LABEL
 "/ ======================================
    start to:stop do:[:aCellIndex|
        self drawLabelAt:aCellIndex x:x y:y h:h.
        y := y + h.
    ].

    y := yTop.

 "/ DRAW INDICATORS
 "/ ===============
    (rendererType == #ComboBox or:[rendererType == #ComboList]) ifTrue:[
        start to:stop do:[:aCellIndex|
            (self hasChoices:aCellIndex) ifTrue:[
                dataSet drawComboButtonAt:aCellIndex x:x y:y w:width
            ].
            y := y + h.
        ]
    ] ifFalse:[
        rendererType == #CheckToggle ifTrue:[
            start to:stop do:[:aCellIndex|
                dataSet drawCheckToggleAt:aCellIndex x:x y:y w:width
                                    state:(self extractColFromRow:(dataSet at:aCellIndex)).
                y := y + h.
            ]
        ]
    ].

 "/ DRAW SEPARATORS
 "/ ===============
    (showColSeparator or:[showRowSeparator]) ifTrue:[
        y    := yTop.
        is3D := dataSet has3Dseparators.

        start to:stop do:[:aCellIndex|
            self drawSeparatorsAt:aCellIndex x:x y:y mode3D:is3D.
            y := y + h.
        ]
    ].

! !

!DataSetColumn methodsFor:'editing'!

editorAt:aRowNr
    |val row|

    row := (dataSet at:aRowNr).
    val := self extractColFromRow:row.

    val isText ifTrue:[val := val string].

    ^ description editorOn:row value:val.
! !

!DataSetColumn methodsFor:'event handling'!

doesNotUnderstand:aMessage

    (description respondsTo:(aMessage selector)) ifTrue:[
        ^ aMessage sendTo:description
    ].
    ^ super doesNotUnderstand:aMessage

! !

!DataSetColumn methodsFor:'grow & degrow'!

growWidth:n
    "grow the width for n pixels
    "
    width := width + n.


!

minWidth
    "get my minimum width required by the entries into the column
    "
    ^ minWidth notNil ifTrue:[minWidth] ifFalse:[self width]


!

setMinWidth
    "set the width to the required width. returns the psitive delta width
    "
    |deltaX|

    deltaX := width - minWidth.
    width  := minWidth.
  ^ deltaX.


! !

!DataSetColumn methodsFor:'initialization'!

on:aDSVColumnView description:aDescription columnNumber:aNumber
    "instance creation; set attributes dependent on the description
    "
    |device selector format idx type oldFont newFont args|

    columnNumber     := aNumber.
    dataSet          := aDSVColumnView.
    description      := aDescription.
    rendererType     := description rendererType.
    width            := nil.
    device           := dataSet device.
    drawableAction   := [:aRow| nil ].
    rowSeparatorSelector := description rowSeparatorSelector.
    fgSelector       := description foregroundSelector.
    bgSelector       := description backgroundSelector.
    showColSeparator := description showColSeparator.
    showRowSeparator := description showRowSeparator.
    columnAdaptor    := dataSet columnAdaptor.
    buttonExtent     := 0 @ 0.
    columnAlignment  := #left.

    (readSelector := description readSelector) notNil ifTrue:[
        numArgsToReadSelector := readSelector numArgs.
    ] ifFalse:[
        numArgsToReadSelector := 0
    ].

    (backgroundColor := description backgroundColor) notNil ifTrue:[
        backgroundColor := backgroundColor on:dataSet device
    ].

    (foregroundColor := description foregroundColor) isNil ifTrue:[
        foregroundColor := dataSet foregroundColor
    ] ifFalse:[
        foregroundColor := foregroundColor on:dataSet device
    ].

    oldFont := nil.
    label   := self resolveLabelWithBuilder:(dataSet builder).

    (label respondsTo:#string) ifTrue:[
        "/
        "/ must set the font to accumulate the real extent of a string label
        "/
        (label isString and:[label isEmpty]) ifTrue:[
            label := nil
        ] ifFalse:[
            (newFont := description labelFont) notNil ifTrue:[
                oldFont := dataSet font.
                dataSet font:newFont.
            ]
        ]
    ].

    oldFont notNil ifTrue:[
        dataSet font:oldFont
    ].

    rendererType == #CheckToggle ifTrue:[
        buttonExtent := dataSet checkToggleExtent.
      ^ self
    ].

    rendererType == #rowSelector ifTrue:[
        buttonExtent := dataSet rowSelectorExtent.
      ^ self
    ].

    (rendererType == #ComboBox or:[rendererType == #ComboList]) ifTrue:[
        buttonExtent := dataSet comboButtonExtent.
    ] ifFalse:[
        columnAlignment := description columnAlignment
    ].

    selector := description printSelector.

    selector notNil ifTrue:[
        args := selector numArgs.
        args == 0 ifTrue:[
            drawableAction := [:aRow| aRow perform:selector ]
        ] ifFalse:[
            args == 1 ifTrue:[
                drawableAction := [:aRow| aRow perform:selector with:dataSet ]
            ] ifFalse:[
                drawableAction := [:aRow| aRow perform:selector with:dataSet with:columnNumber ]
            ]
        ].
        ^ self
    ].

    (     (format := description formatString) notNil
     and:[(type   := description type) == #number or:[type == #numberOrNil]]
    ) ifTrue:[
        "/ has a format string for number (supports only floats)

        (idx := format indexOf:$.) ~~ 0 ifTrue:[
            idx := format size - idx
        ].
        format := '%0.', idx printString, 'f'.

        drawableAction := [:aRow||n|
            n := self extractColFromRow:aRow.
            n isReal ifTrue:[n := n asFloat printfPrintString:format].
            n
        ]
    ] ifFalse:[         "/ default: no format string
        drawableAction := [:aRow| self extractColFromRow:aRow ]
    ]


!

resolveLabelWithBuilder:aBuilder
    |label appl l2|

    aBuilder isNil ifTrue:[
        ^ description rawLabel        
    ].

    (label := description label) isString ifTrue:[
        description labelIsImage ifTrue:[
            aBuilder isEditing ifTrue:[
                label := nil
            ] ifFalse:[
                (label := aBuilder labelFor:(label asSymbol)) notNil ifTrue:[
                    ^ label
                ]
            ].
            ^ description class defaultIcon
        ].
        description translateLabel == true ifTrue:[
            (appl := aBuilder application) notNil ifTrue:[
                (l2 := appl resources string:label) notNil ifTrue:[^ l2]
            ]
        ]
    ].
    ^ label
! !

!DataSetColumn methodsFor:'private'!

extractColFromRow:aRow

    columnAdaptor notNil ifTrue:[
        numArgsToReadSelector == 1 ifTrue:[
            ^ columnAdaptor perform:readSelector with:aRow
        ].
        ^ columnAdaptor perform:readSelector with:aRow with:columnNumber
    ].

    numArgsToReadSelector == 0 ifTrue:[
        ^ aRow perform:readSelector
    ].
    ^ aRow perform:readSelector with:columnNumber
!

storeCol:newValue inRow:aRow
    |writeSelector numArgs|

    writeSelector := description writeSelector.
    numArgs := writeSelector numArgs.

    columnAdaptor notNil ifTrue:[
        numArgs == 2 ifTrue:[
            ^ columnAdaptor perform:writeSelector with:aRow with:newValue
        ].
        ^ columnAdaptor perform:writeSelector with:aRow with:columnNumber with:newValue
    ].

    numArgs == 1 ifTrue:[
        aRow perform:writeSelector with:newValue
    ] ifFalse:[
        aRow perform:writeSelector with:columnNumber with:newValue
    ].
    ^ aRow

! !

!DataSetColumn methodsFor:'queries'!

canSelect:aRowNr
    "returns true if cell in column is selectable
    "
    |s|

    description canSelect ifTrue:[
        (s := description selectSelector) notNil ifTrue:[
            columnAdaptor notNil ifTrue:[
                s numArgs == 2 ifTrue:[
                    ^ columnAdaptor perform:s with:(dataSet at:aRowNr) with:columnNumber
                ].
                ^ columnAdaptor perform:s with:(dataSet at:aRowNr)
            ].
            s numArgs == 1 ifTrue:[
                ^ (dataSet at:aRowNr) perform:s with:columnNumber
            ].
            ^ (dataSet at:aRowNr) perform:s
        ].
        ^ true
    ].
    ^ false

    "Modified: / 7.8.1998 / 22:49:20 / cg"
!

containsText
    "returns true if text might exist
    "
    ^ (rendererType ~~ #rowSelector and:[rendererType ~~ #CheckToggle])
!

hasChoices:aRowNr
    ^ (description choicesFor:(dataSet at:aRowNr)) notNil
!

heightOfHighestRow
    "returns the height of the highest row in pixels
    "
    |hObj hMin end|

    hMin := (dataSet font heightOn:dataSet device) max:(buttonExtent y).

    (hObj := description height) == 0 ifTrue:[
        self containsText ifTrue:[
            "/ search first none empty drawable object
            end := 10 min:dataSet size.

            1 to:end do:[:aRowNr| |lbl|
                lbl := drawableAction value:(dataSet at:aRowNr).

                lbl notNil ifTrue:[
                    (lbl respondsTo:#heightOn:) ifTrue:[
                        hObj := lbl heightOn:dataSet
                    ] ifFalse:[
                        hObj := lbl displayString heightOn:dataSet
                    ].
                    ^ hMin max:hObj
                ]
            ]
        ]
    ].
    ^ hMin max:hObj.
!

showColSeparator
    "returns true if column separator is on
    "
    ^ showColSeparator
! !

!DataSetColumn methodsFor:'searching'!

findRowNrStartingWithChar:aChar start:start stop:stop
    "find the first row starting at start to stop, which drawable label
     starts with the character, aChar. The index of the detected row is
     returned or if no row is found 0.
    "
    |char|

    self containsText ifTrue:[
        char  := aChar asLowercase.

        start to:stop do:[:aRowNr| |lbl|
            lbl := drawableAction value:(dataSet at:aRowNr).

            (lbl respondsTo:#string) ifTrue:[
                lbl := lbl string.
                (lbl size ~~ 0 and:[(lbl at:1) asLowercase == char]) ifTrue:[
                    ^ aRowNr
                ]
            ] ifFalse:[
                lbl isNil ifFalse:[
                    ^ 0
                ]
            ]
        ]
    ].
    ^ 0

! !

!DataSetColumn class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/DataSetColumn.st,v 1.34 1998-09-06 16:10:37 ca Exp $'
! !
